package ch.epfl.ppo.collections;

import java.util.Iterator;
import java.util.NoSuchElementException;

public final class LinkedList<E> implements List<E> {
    private int size;
    private final Node<E> head;

    public LinkedList() {
        Node<E> head = new Node<>(null);
        head.succ = head.pred = head;

        this.size = 0;
        this.head = head;
    }

    @Override
    public boolean isEmpty() {
        return size == 0;
    }

    @Override
    public int size() {
        return size;
    }

    @Override
    public void add(E newElem) {
        Node<E> newNode = new Node<>(head.pred, newElem, head);
        newNode.pred.succ = newNode.succ.pred = newNode;
        size += 1;
    }

    @Override
    public void remove(int i) {
        checkIndex(i);
        removeNode(nodeAtIndex(i));
    }

    @Override
    public E get(int i) {
        checkIndex(i);
        return nodeAtIndex(i).value;
    }

    @Override
    public void set(int i, E newElem) {
        checkIndex(i);
        nodeAtIndex(i).value = newElem;
    }

    @Override
    public Iterator<E> iterator() {
        return new LLIterator<>(this);
    }

    @Override
    public String toString() {
        StringBuilder b = new StringBuilder("[");
        Iterator<E> it = iterator();
        while (it.hasNext()) {
            b.append(it.next());
            if (it.hasNext())
                b.append(",");
        }
        return b.append("]").toString();
    }

    private void checkIndex(int i) {
        if (! (0 <= i && i < size))
            throw new IndexOutOfBoundsException("index out of bounds: " + i);
    }

    private Node<E> nodeAtIndex(int i) {
        Node<E> n = head.succ;
        for (int j = 0; j < i; ++j)
            n = n.succ;
        return n;
    }

    private void removeNode(Node<E> n) {
        n.pred.succ = n.succ;
        n.succ.pred = n.pred;
        size -= 1;
    }

    private static class Node<E> {
        public Node<E> succ, pred;
        public E value;

        public Node(Node<E> pred, E value, Node<E> succ) {
            this.succ = succ;
            this.pred = pred;
            this.value = value;
        }

        public Node(E value) {
            this(null, value, null);
        }
    }

    private static final class LLIterator<E> implements Iterator<E> {
        private final LinkedList<E> list;
        private Node<E> next;
        private boolean canRemove;

        public LLIterator(LinkedList<E> list) {
            this.list = list;
            this.next = list.head.succ;
            this.canRemove = false;
        }

        @Override
        public boolean hasNext() {
            return next != list.head;
        }

        @Override
        public E next() {
            if (! hasNext())
                throw new NoSuchElementException();
            canRemove = true;
            E value = next.value;
            next = next.succ;
            return value;
        }

        @Override
        public void remove() {
            if (! canRemove)
                throw new IllegalStateException();
            list.removeNode(next.pred);
            canRemove = false;
        }
    }
}
